{ Modification made 1998 S Liddicott, search for 'SAMFIX'. samjam@bigfoot.com }
{ Copyright 1995 Tempest Software.  All Rights Reserved.

  Permission is granted to use this software for private
  or commercial use, without fee or royalty.  You may redistribute
  this software as part of an application or DLL without restriction.
  You may also redistribute this software as part of a collection of
  resuable software components, provided all files are included
  without modification.

  This software is released as is, without warranty of any kind,
  either expressed or implied.

  Complete source code is available for $25, payable in US Dollars
  to Tempest Software.  Please send a check to
      Tempest Software
      1125 NW 27th St.
      Corvallis, OR  97330-4337
  Sorry, but we are unable to accept credit cards at this time.
  If you purchase the software, you will also be entitled to
  phone and email technical support.

  If you have an earlier version of the Huge unit, you may continue
  to use it and its source code according to its license (that is,
  free for non-commercial use). You can also use this version of the
  unit, with the new, unrestricted license.  The source code has
  changed little from the earlier version.  We are merely trying
  to create some uniformity for all of our software.

  Please email comments to 'comments@tempest-sw.com'.

  Ray Lischner
  Tempest Software
  17 October 1995
}

unit Huge;

interface

  uses WinTypes, Classes;

  { Class THugeList is similar to TList, but the indexes are of type LongInt,
    rather than Integer.  The size of the list is limited only by the amount
    of virtual memory that can be allocated from Windows.

    Each item is a Pointer.  THugeList never dereferences this pointer, and
    it can be any value you want it to be.  When items are deleted from the
    list and when the list is destroyed, these items are not freed or
    dereferenced.

    List references are always range checked, raising an exception
    for index violations.

    Every effort has been made to ensure that THugeList follows identical
    semantics as TList.

    To make it easier to create derived classes, we use virtual methods for
    the underlying methods, and declare a base class, THugeCustomList, that
    does not define a default array property. THugeList inherits from
    THugeCustomList and defines Items as the default property. Other classes
    derived from THugeCustomList are free to create new properties and make
    one of them the default property.
  }
  type
    THugeCustomList = class
    private
      fHandle: THandle;             { Global memory handle for fList }
      fList: Pointer;               { Pointer to the base of the list }
      fCount: LongInt;              { Number of items in the list }
      fCapacity: LongInt;           { Number of available slots in the list }
    protected
      function GetPointer(Index: LongInt): Pointer;
      procedure ReAllocList(NewCapacity: LongInt);
      procedure CheckIndex(Index: LongInt);
      function ExpandSize: LongInt; virtual;
      function GetItem(Index: LongInt): Pointer;
      procedure SetCapacity(NewCapacity: LongInt);
      procedure SetCount(NewCount: LongInt);
      procedure SetItem(Index: LongInt; Item: Pointer);
      property Handle: THandle read fHandle;
    protected
      property Items[Index: LongInt]: Pointer read GetItem write SetItem;
    public
      destructor Destroy; override;
      function Add(Item: Pointer): LongInt;
      property Capacity: LongInt read fCapacity write SetCapacity;
      procedure Clear; virtual;
      property Count: LongInt read fCount write SetCount;
      procedure Delete(Index: LongInt); virtual;
      procedure Exchange(I1, I2: LongInt); virtual;
      procedure Expand;
      function First: Pointer;
      function IndexOf(Item: Pointer): LongInt; virtual;
      procedure Insert(Index: LongInt; Item: Pointer); virtual;
      function Last: Pointer;
      procedure Move(CurIndex, NewIndex: LongInt); virtual;
      procedure Pack; virtual;
      function Remove(Item: Pointer): LongInt;
    end;
    THugeList = class(THugeCustomList)
    public
      property Items; default;
    end;

    { Implementing a huge string list introduces a dilemma. Do we
      derive the class from TStrings, thereby making it possible, in a limited
      way, to use the huge string list as a Delphi string list, and be
      able to assign it to other TStrings properties, etc., or do we create
      a completely independent class (like THugeCustomList)?  Let's do both.
      THugeStringList is an independent class, with an interface that looks
      and acts like TStringList, but it does not derive from TStrings.

      THugeStrings derives from TStrings, and implements the appropriate
      protocol, but adds a parallel protocol to allow access to the items
      with indexes past 16K. }
    TCharCase = (ccIgnoreCase, ccCaseSensitive);
    type TStrCompareFunc = function(const S1, S2: string): Integer;
    THugeStringList = class(TPersistent)
    private
      fUpdateCount: Cardinal;
      fList: THugeList;
      fDuplicates: TDuplicates;
      fSorted: Boolean;
      fCharCase: TCharCase;
      fOnChange: TNotifyEvent;
      fOnChanging: TNotifyEvent;
    protected
      procedure Changed; virtual;
      procedure Changing; virtual;
      procedure GetCompareFunc(var Func: TStrCompareFunc);
      procedure DefineProperties(Filer: TFiler); override;
      function GetCount: LongInt; virtual;
      function GetItem(Index: LongInt): PStrItem;
      function GetString(Index: LongInt): string; virtual;
      function GetObject(Index: LongInt): TObject; virtual;
      function GetValue(const Name: string): string;
      procedure SetValue(const Name, Value: string);
      function Partition(Lo, Hi: LongInt; Compare: TStrCompareFunc): LongInt; virtual;
      procedure ReadData(Reader: TReader);
      procedure WriteData(Writer: TWriter);
      procedure QuickSort(Lo, Hi: LongInt; Compare: TStrCompareFunc); virtual;
      procedure SlowSort(Lo, Hi: LongInt; Compare: TStrCompareFunc); virtual;
      procedure PutString(Index: LongInt; const S: string); virtual;
      procedure PutObject(Index: LongInt; Obj: TObject); virtual;
      procedure SetSorted(Sorted: Boolean); virtual;
      procedure SetUpdateState(Updating: Boolean); virtual;
      property UpdateCount: Cardinal read fUpdateCount;
      property Items[Index: LongInt]: PStrItem read GetItem;
    public
      constructor Create; virtual;
      destructor Destroy; override;

      function Add(const S: string): LongInt; virtual;
      function AddObject(const S: string; Obj: TObject): LongInt; virtual;
      procedure AddStrings(Strings: THugeStringList); virtual;
      procedure AddTStrings(Strings: TStrings); virtual;
      procedure Assign(Source: TPersistent); virtual;
      procedure BeginUpdate;
      procedure Clear; virtual;
      procedure Delete(Index: LongInt); virtual;
      procedure EndUpdate;
      procedure Exchange(Index1, Index2: LongInt); virtual;
      function Find(const S: string; var Index: LongInt): Boolean; virtual;
      function GetText: PChar; virtual;
      function IndexOf(const S: string): LongInt; virtual;
      function IndexOfObject(Obj: TObject): LongInt;
      procedure Insert(Index: LongInt; const S: string); virtual;
      procedure InsertObject(Index: LongInt; const S: string; Obj: TObject);
      procedure LoadFromFile(const FileName: string);
      procedure LoadFromStream(Stream: TStream); virtual;
      procedure Move(CurIndex, NewIndex: LongInt); virtual;
      procedure SaveToFile(const FileName: string);
      procedure SaveToStream(Stream: TStream); virtual;
      procedure SetText(Text: PChar); virtual;
      procedure Sort; virtual;

      property CharCase: TCharCase read fCharCase write fCharCase;
      property Count: LongInt read GetCount;
      property Duplicates: TDuplicates read fDuplicates write fDuplicates;
      property Sorted: Boolean read FSorted write SetSorted;
      property Objects[Index: LongInt]: TObject read GetObject write PutObject;
      property Strings[Index: LongInt]: string read GetString write PutString; default;
      property Values[const Name: string]: string read GetValue write SetValue;
      property OnChange: TNotifyEvent read fOnChange write fOnChange;
      property OnChanging: TNotifyEvent read fOnChanging write fOnChanging;
    end;

    THugeStrings = class(TStrings)
    private
      fList: THugeStringList;
      function GetDuplicates: TDuplicates;
      function GetSorted: Boolean;
      function GetOnChange: TNotifyEvent;
      function GetOnChanging: TNotifyEvent;
      procedure SetOnChange(OnChange: TNotifyEvent);
      procedure SetOnChanging(OnChanging: TNotifyEvent);
      procedure SetDuplicates(Duplicates: TDuplicates);
      procedure SetSorted(Sorted: Boolean);
    protected
      function Get(Index: Integer): string; override;
      function GetCount: Integer; override;
      function HugeGetCount: LongInt; virtual;
      function GetObject(Index: Integer): TObject; override;
      procedure Put(Index: Integer; const S: string); override;
      procedure PutObject(Index: Integer; Obj: TObject); override;
      procedure SetUpdateState(Updating: Boolean); override;
      function HugeGet(Index: LongInt): string; virtual;
      function HugeGetObject(Index: LongInt): TObject; virtual;
      procedure HugePut(Index: LongInt; const S: string); virtual;
      procedure HugePutObject(Index: LongInt; Obj: TObject); virtual;
      function HugeGetValue(const Name: string): string; virtual;
      procedure HugeSetValue(const Name, Value: string); virtual;
      procedure ReadData(Reader: TReader); virtual;
      procedure WriteData(Writer: TWriter); virtual;
      procedure DefineProperties(Filer: TFiler); override;
      function Trunc(Index: LongInt): Integer;
    public
      constructor Create;
      destructor Destroy; override;

      function Add(const S: string): Integer; override;
      function HugeAdd(const S: string): LongInt; virtual;
      function AddObject(const S: string; Obj: TObject): Integer; override;
      function HugeAddObject(const S: string; Obj: TObject): LongInt; virtual;
      procedure Assign(Source: TPersistent); override;
      procedure Clear; override;
      procedure Delete(Index: Integer); override;
      procedure HugeDelete(Index: LongInt); virtual;
      procedure Exchange(Index1, Index2: Integer); override;
      procedure HugeExchange(Index1, Index2: LongInt); virtual;
      function Find(const S: string; var Index: Integer): Boolean; virtual;
      function HugeFind(const S: string; var Index: LongInt): Boolean; virtual;
      function GetText: PChar; override;
      function IndexOf(const S: string): Integer; override;
      function HugeIndexOf(const S: string): LongInt; virtual;
      function HugeIndexOfObject(Obj: TObject): LongInt; virtual;
      procedure Insert(Index: Integer; const S: string); override;
      procedure HugeInsert(Index: LongInt; const S: string); virtual;
      procedure HugeInsertObject(Index: LongInt; const S: string; Obj: TObject);
      procedure LoadFromStream(Stream: TStream); override;
      procedure Move(CurIndex, NewIndex: Integer); override;
      procedure HugeMove(CurIndex, NewIndex: LongInt); virtual;
      procedure SaveToStream(Stream: TStream); override;
      procedure SetText(Text: PChar); override;
      procedure Sort; virtual;

      property Count: Integer read GetCount;
      property HugeCount: LongInt read HugeGetCount;
      property Duplicates: TDuplicates read GetDuplicates write SetDuplicates;
      property Sorted: Boolean read GetSorted write SetSorted;
      property HugeObjects[Index: LongInt]: TObject read HugeGetObject write HugePutObject;
      property HugeStrings[Index: LongInt]: string read HugeGet write HugePut;
      property HugeValues[const Name: string]: string read HugeGetValue write HugeSetValue;
      property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
      property OnChanging: TNotifyEvent read GetOnChanging write SetOnChanging;
    end;

  { HugeInc() and HugeOffset() add an offset to a far pointer.
    The offset can be greater than 64K, and the pointer's segment
    is properly updated.  Both a procedure and a function are
    provided; they do the same thing, but return the result
    differently.  Choose whichever is more convenient.
  }
  procedure HugeInc(var HugePtr: Pointer; Amount: LongInt);
  procedure HugeDec(var HugePtr: Pointer; Amount: LongInt);
  function  HugeOffset(HugePtr: Pointer; Amount: LongInt): Pointer;

  { CopyLine copies the string S, to Text, and then appends
    a CR-LF (Carriage Return-Line Feed) for a line ending.
    It returns a pointer just past the LF, in preparation
    for copying the next line.  The memory pointed to by
    Text can be larger than 64K. }
  function CopyLine(Text: PChar; const S: string): PChar;
  { ExtractLine copies a line of text from Text to S. A line
    ends with a CR-LF, a plain CR or a plain LF, or the end
    of the string, which can be marked by a Ctrl+Z or a zero
    byte.  It returns a pointer to the beginning of the next
    line, or the Ctrl+Z or zero terminating character.
    The memory pointed to by Text can be larger than 64K. }
  function ExtractLine(Text: PChar; var S: string): PChar;

  procedure Register;

  { The threshold between QuickSort and SlowSort. Using SlowSort
    for small sorts is faster. The number was determined empirically
    using case-insensitive sorts of small strings on a Pentium.
    Your situation may vary, so you can adjust the treshold to meet your needs. }
  const
    SortThreshold: Integer = 16;

implementation

  uses Consts, SysUtils, WinProcs;

  const
    NUL=#0;
    LF=#10;
    CR=#13;
    EOF=#26;

  procedure HugeShift; far; external 'KERNEL' index 113;

  procedure HugeInc(var HugePtr: Pointer; Amount: LongInt); assembler;
  asm
    mov ax, Amount.Word[0]  { Store Amount in DX:AX }
    mov dx, Amount.Word[2]
    les bx, HugePtr         { Get the reference to HugePtr }
    add ax, es:[bx]         { Add the offset parts }
    adc dx, 0               { Propagate carry to the high word of Amount }
    mov cx, OFFSET HugeShift
    shl dx, cl              { Shift high word of Amount for segment }
    add es:[bx+2], dx       { Increment the segment of HugePtr }
    mov es:[bx], ax
  end;
{ Pascal equivalent:
  Inc(Amount, LongRec(HugePtr).Lo);
  Inc(LongRec(HugePtr).Hi, SelectorInc * LongRec(Amount).Hi);
  LongRec(HugePtr).Lo := LongRec(Amount).Lo;
}

  procedure HugeDec(var HugePtr: Pointer; Amount: LongInt); assembler;
  asm
    les bx, HugePtr         { Store HugePtr ptr in es:[bx] }
    mov ax, es:[bx]
    sub ax, Amount.Word[0]  { Subtract the offset parts }
    mov dx, Amount.Word[2]
    adc dx, 0               { Propagate carry to the high word of Amount }
    mov cx, OFFSET HugeShift
    shl dx, cl              { Shift high word of Amount for segment }
    sub es:[bx+2], dx
    mov es:[bx], ax
  end;

  function HugeOffset(HugePtr: Pointer; Amount: LongInt): Pointer; assembler;
  asm
    mov ax, Amount.Word[0]  { Store Amount in DX:AX }
    mov dx, Amount.Word[2]
    add ax, HugePtr.Word[0] { Add the offset parts }
    adc dx, 0               { Propagate carry to the high word of Amount }
    mov cx, OFFSET HugeShift
    shl dx, cl              { Shift high word of Amount for segment }
    add dx, HugePtr.Word[2] { Increment the segment of HugePtr }
  end;

  { Having written the generic routines, HugeOffset, etc., we do not
    use them in THugeCustomList.  Instead, THugeCustomList relies on the
    fact that Windows always returns a segment-aligned pointer from
    GlobalAlloc. Thus, we know the memory base has a zero offset, so we
    can further optimize our pointer references.  This optimization cuts
    off about 1/3 of the run-time on a Pentium 75.

    I tried further optimization by making FastHugeOffset into an inline
    assembler function, but that wasn't any faster.  Hence, let's stick
    with the readability of Pascal.
  }
  function FastHugeOffset(Segment: Word; Offset: LongInt): Pointer;
  begin
    Result := Ptr(Segment + LongRec(Offset).HI shl Ofs(HugeShift), LongRec(Offset).Lo);
  end;

  {
    Copy memory from Src to Dst, copying Size units from
    Base+Src to Base+Dst.  The reason we separate the Base pointer
    from the Dst and Src indexes is to we can easily compare
    the indexes to see what direction we are copying, without
    having to muck about with normalizing segments.
    Offsets and the size are in longwords, for our convenience,
    but we need to convert them to short words, in case we are
    running on a 286 machine, which does not support the movsd
    instruction, just movsw.

    I'm not a fan of writing assembly language, so most of the code
    is Pascal, and I only dip into assembly for the low-level
    copying code, in DoMove, plus setting the direction flag,
    which we do with inline assembler procedures.
  }
  procedure cld; inline ($fc);  { clear direction flag }
  procedure std; inline ($fd);  { set direction flag }

  procedure DoMove(SrcPtr, DstPtr: Pointer; Size: Word); assembler;
  asm
    push ds
    lds si, SrcPtr
    les di, DstPtr
    mov cx, Size.Word[0]
    rep movsw
    pop ds
  end;

  { When setting the Count, we might add many new items, which we
    want to zero at one time, to initialize all items to nil.
    We could use FillChar, which fills by bytes, but, as DoMove
    is to Move, ZeroBytes is to FillChar, except that we always
    fill with zero words. }
  procedure ZeroWords(DstPtr: Pointer; Size: Word); assembler;
  asm
    xor ax, ax
    les di, DstPtr
    mov cx, Size.Word[0]
    cld
    rep stosw
  end;

  { When computing how many words to move, we need to see
    how far from a segment boundary SrcPtr and DstPtr are,
    since the MOVS instruction won't cross a segment boundary
    properly.  Since we know that the list is allocated on
    a segment boundary, and we store long words, we know that
    SrcPtr and DstPtr will always be on longword boundaries.

    It is actually a little tricky to determine how many words
    we can copy before hitting a segment boundary.  It depends on
    whether we are copying up or down.  When copying down, the
    difference between $10000 and the pointer offset determines
    how many bytes we can copy in that segment.  If the segment
    offset is zero, then truncating $10000 to a word yields zero,
    in which case we copy the maximum number of words, $7FFF.
  }
  function ComputeDownMoveSize(SrcOfs, DstOfs: Word): Word;
  begin
    if SrcOfs > DstOfs then
      Result := Word($10000 - SrcOfs) div 2
    else
      Result := Word($10000 - DstOfs) div 2;
    if Result = 0 then
      Result := $7FFF;
  end;
  function ComputeUpMoveSize(SrcOfs, DstOfs: Word): Word;
  begin
    if SrcOfs = $FFFF then
      Result := DstOfs div 2
    else if DstOfs = $FFFF then
      Result := SrcOfs div 2
    else if SrcOfs > DstOfs then
      Result := DstOfs div 2 + 1
    else
      Result := SrcOfs div 2 + 1;
  end;

  procedure HugeMove(Base: Pointer; Dst, Src, Size: LongInt); far;
  var
    SrcPtr, DstPtr: Pointer;
    MoveSize: Word;
  begin
    SrcPtr := HugeOffset(Base, Src * SizeOf(Pointer));
    DstPtr := HugeOffset(Base, Dst * SizeOf(Pointer));

    { Convert longword size to words. }
    Size := Size * (SizeOf(LongInt) div SizeOf(Word));

    if Src < Dst then
    try {SAMFIX Sam Liddicott 1998, Add try-finally MUST do a CLD afterwards, at ALL COSTS }
      { start from the far end and work towards the front }
      std;
      HugeInc(SrcPtr, (Size-1) * SizeOf(Word));
      HugeInc(DstPtr, (Size-1) * SizeOf(Word));

      while Size > 0 do
      begin
        MoveSize := ComputeUpMoveSize(Word(SrcPtr), Word(DstPtr));
        if MoveSize > Size then
          MoveSize := Word(Size);
        DoMove(SrcPtr, DstPtr, MoveSize);
        Dec(Size, MoveSize);
        HugeDec(SrcPtr, MoveSize * SizeOf(Word));
        HugeDec(DstPtr, MoveSize * SizeOf(Word));
      end
    finally
      cld; {SAMFIX, add CLD, Some API functions fail if the direction flag is wrong.  }
    end
    else
    begin
      cld;
      while Size > 0 do
      begin
        MoveSize := ComputeDownMoveSize(Word(SrcPtr), Word(DstPtr));
        if MoveSize > Size then
          MoveSize := Word(Size);
        DoMove(SrcPtr, DstPtr, MoveSize);
        Dec(Size, MoveSize);
        HugeInc(SrcPtr, MoveSize * SizeOf(Word));
        HugeInc(DstPtr, MoveSize * SizeOf(Word));
      end;
    end;
  end;

  { Destroy a THugeCustomList by unlocking and freeing the list's memory }
  destructor THugeCustomList.Destroy;
  begin
    if Handle <> 0 then
    begin
      GlobalUnlock(Handle);
      GlobalFree(Handle);
    end;
    inherited Destroy;
  end;

  { Reallocate the list, copying the old list values into
    the new list.  Reset all the fields for the new list.
  }
  procedure THugeCustomList.ReAllocList(NewCapacity: LongInt);
  var
    NewHandle: THandle;
  begin
    if Handle = 0 then
      NewHandle := GlobalAlloc(GPTR, NewCapacity*SizeOf(Pointer))
    else
      NewHandle := GlobalReAlloc(Handle, NewCapacity*SizeOf(Pointer), GMEM_ZEROINIT);

    if NewHandle = 0 then
      OutOfMemoryError;
    try
      fList := GlobalLock(NewHandle);
      if fList = nil then
        OutOfMemoryError;

      { If the list shrunk, then update Count for the smaller size. }
      if NewCapacity < Count then
        fCount := NewCapacity;

      { If Windows allocated even more memory, then see how much
        we really have.  The user might not have requested that
        much, but why waste memory?
      }
      fCapacity := GlobalSize(NewHandle) div SizeOf(Pointer);

      { And remember all the new information }
      fHandle := NewHandle;
    except
      { Oops. Something went wrong, make sure we don't leave the
        new memory dangling.  If something goes really, really
        wrong, and the new memory has already been stored in Self,
        then leave it alone.
      }
      if Handle <> NewHandle then
      begin
        GlobalUnlock(NewHandle);
        GlobalFree(NewHandle);
      end;
      raise;
    end;
  end;

  { Return the new capacity of the list when we expand it.
    Here we emulate TList, which expands the list by 4, 8, or
    16 elements, depending on the number of items already in
    the list.
  }
  function THugeCustomList.ExpandSize: LongInt;
  begin
    if Capacity > 8 then
      ExpandSize := Capacity + 16
    else if Capacity > 4 then
      ExpandSize := Capacity + 8
    else
      ExpandSize := Capacity + 4
  end;

  { Check a list index and raise an exception if it is not valid }
  procedure THugeCustomList.CheckIndex(Index: LongInt);
  begin
    if (Index < 0) or (Index >= Count) then
      raise EListError.CreateRes(SListIndexError)
  end;

  { Return an item of the list, raising an exception for an invalid index }
  function THugeCustomList.GetItem(Index: LongInt): Pointer;
  var
    Ptr: ^Pointer;
  begin
    CheckIndex(Index);
    Ptr := FastHugeOffset(LongRec(fList).Hi, Index * SizeOf(Pointer));
    GetItem := Ptr^;
  end;

  { Get an item without checking for range errors.  Return a pointer
    to the slot in the list, so the pointer value can be changed.
    This function is available for use by derived classes, but is
    not used by THugeCustomList itself.  Instead, THugeCustomList calls
    FastHugeOffset() directly, which improves performance.
  }
  function THugeCustomList.GetPointer(Index: LongInt): Pointer;
  begin
    GetPointer := FastHugeOffset(LongRec(fList).Hi, Index * SizeOf(Pointer));
  end;

  { Set a list item. }
  procedure THugeCustomList.SetItem(Index: LongInt; Item: Pointer);
  var
    Ptr: ^Pointer;
  begin
    CheckIndex(Index);
    Ptr := FastHugeOffset(LongRec(fList).Hi, Index * SizeOf(Pointer));
    Ptr^ := Item;
  end;

  { Set the new capacity of the list.  If the list shrinks,
    then adjust Count.
  }
  procedure THugeCustomList.SetCapacity(NewCapacity: LongInt);
  begin
    if NewCapacity < 0 then
      raise EListError.CreateRes(sListIndexError);
    if Capacity <> NewCapacity then
      ReAllocList(NewCapacity);
  end;

  { Set the count. If the new count is larger, then set the new
    items to nil. Increase the capacity as needed. }
  procedure THugeCustomList.SetCount(NewCount: LongInt);
  var
    Ptr: ^Pointer;
    NZeroBytes: LongInt;
    NBytes: Word;
  begin
    if NewCount < 0 then
      raise EListError.CreateRes(sListIndexError);
    if NewCount > Capacity then
      Capacity := NewCount;
    if Count < NewCount then
    begin
      { zero all the new items, which may cross segment boundaries }
      Ptr := FastHugeOffset(LongRec(fList).Hi, Count * SizeOf(Pointer));
      NZeroBytes := (NewCount - Count) * SizeOf(Pointer);
      while NZeroBytes > 0 do
      begin
        if Ofs(Ptr^) = 0 then
          NBytes := $FFFE
        else
          NBytes := $10000 - Ofs(Ptr^);
        if NBytes > NZeroBytes then
          NBytes := NZeroBytes;
        ZeroWords(Ptr, NBytes div 2);
        Dec(NZeroBytes, NBytes);
        Ptr := HugeOffset(Ptr, NBytes);
      end;
    end;
    fCount := NewCount;
  end;

  { Add Item to the end of the list }
  function THugeCustomList.Add(Item: Pointer): LongInt;
  begin
    Insert(Count, Item);
    Add := Count-1;
  end;

  { Clear all the items in the list. }
  procedure THugeCustomList.Clear;
  begin
    if Handle <> 0 then
    begin
      GlobalUnlock(Handle);
      GlobalFree(Handle);
      fHandle := 0;
      fList := nil;
    end;
    fCount := 0;
    fCapacity := 0;
  end;

  { Delete the item at Index, shifting down all other items
    with higher indexes.
  }
  procedure THugeCustomList.Delete(Index: LongInt);
  begin
    CheckIndex(Index);
    Dec(fCount);
    HugeMove(fList, Index, Index+1, Count-Index);
  end;

  { Exchange the items at indexes I1 and I2. }
  procedure THugeCustomList.Exchange(I1, I2: LongInt);
  var
    Tmp: Pointer;
    P, Q: ^Pointer;
  begin
    CheckIndex(I1);
    CheckIndex(I2);
    P := FastHugeOffset(LongRec(fList).Hi, I1 * SizeOf(Pointer));
    Q := FastHugeOffset(LongRec(fList).Hi, I2 * SizeOf(Pointer));
    Tmp := P^;
    P^ := Q^;
    Q^ := Tmp;
  end;

  { Expand the list's capacity to make room for more items.
    This is done automatically when new items are added,
    so there is little reason for the user to call Expand
    explicitly.
  }
  procedure THugeCustomList.Expand;
  begin
    Capacity := ExpandSize;
  end;

  { Return the first item in the list }
  function THugeCustomList.First: Pointer;
  begin
    First := Items[0];
  end;

  { Return the index of the first occurrence of Item.
    Return -1 for not found.
  }
  function THugeCustomList.IndexOf(Item: Pointer): LongInt;
  var
    I: LongInt;
    Ptr: ^Pointer;
  begin
    IndexOf := -1;
    for I := 0 to Count-1 do
    begin
      Ptr := FastHugeOffset(LongRec(fList).Hi, I * SizeOf(Pointer));
      if Ptr^ = Item then
      begin
        IndexOf := I;
        Break;
      end;
    end;
  end;

  { Insert Item at position, Index.  Slide all other items
    over to make room.  The user can insert to any valid index,
    or to one past the end of the list, thereby appending an
    item to the list.  In the latter case, adjust the capacity
    if needed.
  }
  procedure THugeCustomList.Insert(Index: LongInt; Item: Pointer);
  var
    i: LongInt;
    Ptr: ^Pointer;
    NextPtr: ^Pointer;
  begin
    if (Index < 0) or (Index > Count) then
      raise EListError.CreateRes(SListIndexError);
    if Count >= Capacity then
      Expand;

    { Make room for the inserted item. }
    Ptr := FastHugeOffset(LongRec(fList).Hi, Index * SizeOf(Pointer));
    HugeMove(Ptr, 1, 0, Count-Index);

    Ptr^ := Item;
    Inc(fCount);
  end;

  { Return the last item in the list.  Raise an exception
    if the list is empty.
  }
  function THugeCustomList.Last: Pointer;
  begin
    Last := Items[Count - 1];
  end;

  { Move an item from CurIndex to NewIndex.  We only need to move
    the items that lie between CurIndex and NewIndex, leaving the
    rest of the list alone.
  }
  procedure THugeCustomList.Move(CurIndex, NewIndex: LongInt);
  var
    Tmp: Pointer;
    Ptr: ^Pointer;
  begin
    CheckIndex(NewIndex);
    if NewIndex <> CurIndex then
    begin
      Tmp := Items[CurIndex];
      if NewIndex < CurIndex then
      begin
        Ptr := FastHugeOffset(LongRec(fList).Hi, NewIndex * SizeOf(Pointer));
        HugeMove(Ptr, 1, 0, CurIndex-NewIndex)
      end
      else if CurIndex < NewIndex then
      begin
        Ptr := FastHugeOffset(LongRec(fList).Hi, CurIndex * SizeOf(Pointer));
        HugeMove(Ptr, 0, 1, NewIndex-CurIndex);
        Ptr := FastHugeOffset(LongRec(fList).Hi, NewIndex * SizeOf(Pointer));
      end;
      Ptr^ := Tmp;
    end;
  end;

  { Pack the list by removing nil slots.  After packing
    Count might be smaller.  After each loop iteration,
    we keep invariant:
      Items[k] <> nil for all k <= i
    Thus, when we reach the end of the loop, then we know we have
    packed the list.  We march through the list, using the I index.
    Whenever Items[i] = nil, we collect a maximal string of nil slots,
    and then shift down the remaining items, adjusting Count to match.
  }
  procedure THugeCustomList.Pack;
  var
    I, J, K: LongInt;
    P, Q: ^Pointer;
  begin
    { Instead of a for loop, use a while loop, so we use the
      current value of Count for each iteration, since Count
      changes during the loop. }
    I := 0;
    P := fList;
    while I < Count do
    begin
      if P^ <> nil then
      begin
        Inc(I);
        P := FastHugeOffset(LongRec(fList).Hi, I*SizeOf(Pointer));
      end
      else
      begin
        { Collect a run of nil slots }
        for J := I+1 to Count-1 do
        begin
          P := FastHugeOffset(LongRec(fList).Hi, J * SizeOf(Pointer));
          if P^ <> nil then
            Break;
        end;
        { We only need to shift slots if there is a non-nil value.
          If all the remaining slots are nil, then we are done.
        }
        if P^ = nil then
        begin
          fCount := i;
          Break;
        end;

        { Now shift the slots; setting the newly vacated slots to nil, just
          in case we accidentally try to refer to one of them.
          Stop when we get to a nil slot.
        }
        K := I;
        while J < Count do
        begin
          P := FastHugeOffset(LongRec(fList).Hi, K * SizeOf(Pointer));
          Q := FastHugeOffset(LongRec(fList).Hi, J * SizeOf(Pointer));
          P^ := Q^;
          { Check after assigning to P^, so the check for nil at
            the top of the loop is true.  A small inefficiency for
            greater programming ease and maintainability.
          }
          if Q^ = nil then
            Break;
          Q^ := nil;
          Inc(K);
          Inc(J);
        end;
        { Adjust Count by the number of nil slots removed }
        Dec(fCount, J-K);
        { Set the loop counter to the next nil slot }
        I := K;
      end;
    end;
  end;

  { Remove Item and return its index. }
  function THugeCustomList.Remove(Item: Pointer): LongInt;
  begin
    Result := IndexOf(Item);
    if Result <> -1 then
      Delete(Result);
  end;


  { THugeStringList }
  { Copy the string, S, to Text, and append a CR-LF.
    Even though we are copying no more than 256 bytes, we use
    HMemCopy, since Text might cross a segment boundary.
  }
  function CopyLine(Text: PChar; const S: string): PChar;
  var
    Index, N: Integer;
  begin
    HMemCpy(Text, @s[1], Length(S));
    Result := HugeOffset(Text, Length(S));
    { Now append the CR-LF }
    Result^ := CR;
    Result := HugeOffset(Result, 1);
    Result^ := LF;
    Result := HugeOffset(Result, 1);
  end;

  { Extract the next line from Text, and copy it into S.
    Advance Text past the end of the string. A string ends with
    a line ending (CRLF, CR, or LF), at the end of the string
    (Ctrl+Z or zero byte), or at 255 bytes. A string can cross
    a segment boundary (which makes for the really ugly business).
    The line ending is not included in the string, S. }
  function ExtractLine(Text: PChar; var S: string): PChar;
  var
    Len: Integer;
  begin
    Len := 0;
    Result := Text;
    while not (Result^ in [NUL, LF, CR, EOF]) do
    begin
      Result := HugeOffset(Result, 1);
      Inc(Len);
      if Len = 255 then
        Break;
    end;
    S[0] := Chr(Len);
    HMemCpy(@S[1], Text, Len);
    if Result^ = CR then
      Result := HugeOffset(Result, 1);
    if Result^ = LF then
      Result := HugeOffset(Result, 1);
  end;

  constructor THugeStringList.Create;
  begin
    inherited Create;
    fList := THugeList.Create;
  end;

  destructor THugeStringList.Destroy;
  begin
    Clear;
    fList.Free;
    inherited Destroy;
  end;

  function THugeStringList.GetCount: LongInt;
  begin
    Result := fList.Count
  end;

  function THugeStringList.GetString(Index: LongInt): string;
  begin
    Result := PStrItem(fList[Index])^.fString
  end;

  function THugeStringList.GetObject(Index: LongInt): TObject;
  begin
    Result := PStrItem(fList[Index])^.fObject
  end;

  { Since the TStrItem has only enough room for the string it holds,
    we need to allocate a brand new one. We save ourselves the trouble
    when Length(newstr)=Length(oldstr). Either way, we need to copy the
    new string, and this way, we save ourselves a dynamic allocation. }
  procedure THugeStringList.PutString(Index: LongInt; const S: string);
  var
    Old: PStrItem;
  begin
    if Sorted then
      raise EListError.CreateRes(SSortedListError);
    Old := fList[Index];
    if Length(S) = Length(Old^.fString) then
      Old^.fString := S
    else
    begin
      fList[Index] := NewStrItem(S, Old^.fObject);
      DisposeStrItem(Old);
    end;
  end;

  procedure THugeStringList.PutObject(Index: LongInt; Obj: TObject);
  begin
    PStrItem(fList[Index])^.fObject := Obj
  end;

  { An unnested BeginUpdate sets the update state, and when
    the nesting returns to zero in EndUpdate, that means
    we are done making changes. }
  procedure THugeStringList.SetUpdateState(Updating: Boolean);
  begin
    if Updating then
      Changing
    else
      Changed
  end;

  { The list has changed. }
  procedure THugeStringList.Changed;
  begin
    if Assigned(fOnChange) then
      fOnChange(Self)
  end;

  { The list is about to change. }
  procedure THugeStringList.Changing;
  begin
    if Assigned(fOnChanging) then
      fOnChanging(Self)
  end;

  { Clearing the Sorted flag leaves the order alone. Setting
    the Sorted flag immediately sorts the list. }
  procedure THugeStringList.SetSorted(Sorted: Boolean);
  begin
    if Sorted <> fSorted then
    begin
      if Sorted then
        Sort;
      fSorted := Sorted;
    end;
  end;

  { Add a string, allocating room for it. If the list is sorted,
    and the string is already in the list, then what we do
    depends on the Duplicates property. }
  function THugeStringList.Add(const S: string): LongInt;
  begin
    if not Sorted then
      Result := FList.Count
    else
      if Find(S, Result) then
        case Duplicates of
          dupIgnore: Exit;
          dupError: raise EListError.CreateRes(SDuplicateString);
          dupAccept: ; { nothing, so we can insert the string below }
        end;
    Changing;
    fList.Insert(Result, NewStrItem(S, nil));
    Changed;
  end;

  { Add a string and an associated object. }
  function THugeStringList.AddObject(const S: string; Obj: TObject): LongInt;
  begin
    Result := Add(S);
    Objects[Result] := Obj;
  end;

  { Add all the strings in the THugeStringList. }
  procedure THugeStringList.AddStrings(Strings: THugeStringList);
  var
    I: LongInt;
  begin
   BeginUpdate;
   try
     for I := 0 to Strings.Count-1 do
       AddObject(Strings.Strings[I], Strings.Objects[I]);
   finally
     EndUpdate;
   end;
  end;

  { Add all the strings in a TStrings. This makes it convenient
    to copy the strings from, say, a TMemo or TListBox into a
    THugeStringList. }
  procedure THugeStringList.AddTStrings(Strings: TStrings);
  var
    I: Integer;
  begin
   BeginUpdate;
   try
     for I := 0 to Strings.Count-1 do
       AddObject(Strings.Strings[I], Strings.Objects[I]);
   finally
     EndUpdate;
   end;
  end;

  { Accept assignments from a TStrings, for the user's convenience, but
    if the TStrings is a THugeStrings, then copy all the strings
    by accessing the THugeStringList field. }
  procedure THugeStringList.Assign(Source: TPersistent);
  begin
    if Source is THugeStrings then
      Assign(THugeStrings(Source).fList)
    else if Source is TStrings then
    begin
      BeginUpdate;
      try
        Clear;
        AddTStrings(TStrings(Source));
      finally
        EndUpdate;
      end;
    end
    else if Source is THugeStringList then
    begin
      BeginUpdate;
      try
        Clear;
        AddStrings(THugeStringList(Source));
      finally
        EndUpdate;
      end;
    end
    else
      inherited Assign(Source);
  end;

  procedure THugeStringList.BeginUpdate;
  begin
    if fUpdateCount = 0 then
      SetUpdateState(True);
    Inc(fUpdateCount);
  end;

  procedure THugeStringList.Clear;
  var
    I: LongInt;
  begin
    Changing;
    for I := 0 to fList.Count-1 do
      DisposeStrItem(fList[I]);
    fList.Clear;
    Changed;
  end;

  { Set the comparison function, depending on the CharCase property. }
  procedure THugeStringList.GetCompareFunc(var Func: TStrCompareFunc);
  begin
    case CharCase of
    ccCaseSensitive: Func := AnsiCompareStr;
    ccIgnoreCase:    Func := AnsiCompareText;
    end;
  end;

  { Define a pseudo-property, so we can save a huge string list in a form file. }
  procedure THugeStringList.DefineProperties(Filer: TFiler);
  begin
    Filer.DefineProperty('Strings', ReadData, WriteData, Count>0);
  end;

  { Read strings from a TReader, e.g., a form file. }
  procedure THugeStringList.ReadData(Reader: TReader);
  begin
    Reader.ReadListBegin;
    BeginUpdate;
    try
      Clear;
      while not Reader.EndOfList do
        Add(Reader.ReadString);
      Reader.ReadListEnd;
    finally
      EndUpdate;
    end;
  end;

  { Write strings to a TWriter, e.g., a form file }
  procedure THugeStringList.WriteData(Writer: TWriter);
  var
    I: LongInt;
  begin
    Writer.WriteListbegin;
    for I := 0 to Count-1 do
      Writer.WriteString(Items[I]^.fString);
    Writer.WriteListEnd;
  end;

  procedure THugeStringList.Delete(Index: LongInt);
  begin
    Changing;
    DisposeStrItem(fList[Index]);
    fList.Delete(Index);
    Changed;
  end;

  procedure THugeStringList.EndUpdate;
  begin
    Dec(fUpdateCount);
    if fUpdateCount = 0 then
      SetUpdateState(False);
  end;

  { Exchange the items at Index1 and Index2. Clearly, we can't
    do this if the list is sorted. }
  procedure THugeStringList.Exchange(Index1, Index2: LongInt);
  begin
    if Sorted then
      raise EListError.CreateRes(SSortedListError);
    Changing;
    fList.Exchange(Index1, Index2);
    Changed;
  end;

  { Use binary search to find something in a sorted list. If the list
    is unsorted, then switch to IndexOf. }
  function THugeStringList.Find(const S: string; var Index: LongInt): Boolean;
  var
    Lo, Hi, Mid: LongInt;
    C: Integer;
    Compare: TStrCompareFunc;
  begin
    if not Sorted then
    begin
      Index := IndexOf(S);
      Result := Index >= 0;
    end
    else
    begin
      Result := False;
      Lo := 0;
      Hi := Count-1;
      GetCompareFunc(Compare);
      while Lo <= Hi do
      begin
        Mid := (Lo + Hi) div 2;
        C := Compare(Items[Mid]^.fString, S);
        if C < 0 then
          Lo := Mid + 1
        else
        begin
          Hi := Mid - 1;
          if C = 0 then
          begin
            Result := True;
            if Duplicates <> dupAccept then
              Lo := Mid;
          end;
        end;
      end;
      Index := Lo;
    end;
  end;

  { Get a string item. This is just a convenience, to make the
    code more readable than using explicit type casts. }
  function THugeStringList.GetItem(Index: LongInt): PStrItem;
  begin
    Result := fList[Index]
  end;

  { Copy all the strings into one, big string, ending each
    string with a CR-LF.  End the entire string with a zero byte.
    Since this is a Huge string list, we allow the resulting string
    to be any size, so long as we can allocate enough memory for it. }
  function THugeStringList.GetText: PChar;
  var
    I, Size: LongInt;
    P: PChar;
    Handle: THandle;
  begin
    Size := 1;           { for the trailing #0 byte }
    for I := 0 to Count-1 do
      Size := Size + Length(Items[I]^.fString) + 2;  { CR-LF }
    Handle := GlobalAlloc(GMem_Fixed or GMem_Share, Size);
    if Handle = 0 then
      OutOfMemoryError;
    Result := GlobalLock(Handle);
    if Result = nil then
    begin
      GlobalFree(Handle);
      OutOfMemoryError;
    end;
    P := Result;
    for I := 0 to Count-1 do
      P := CopyLine(P, Items[I]^.fString);
    P[0] := #0;
  end;

  { Get a value from a string. String/value pairs are expressed
    like environment variables, e.g., Name=Value.  Look for the
    equal sign, and copy everything that comes after it. }
  function THugeStringList.GetValue(const Name: string): string;
  var
    Equal: Integer;
    I: LongInt;
    Compare: TStrCompareFunc;
  begin
    GetCompareFunc(Compare);
    for I := 0 to Count-1 do
    begin
      Equal := Pos('=', Items[I]^.fString);
      if (Equal <> 0) and (Compare(Name, Copy(Items[I]^.fString, 1, Equal-1)) = 0) then
      begin
        Result := Copy(Items[I]^.fString, Equal+1, 255);
        Exit;
      end;
    end;
    Result := '';
  end;

  { Set the value of a string/value pair. Look for the equal sign,
    and change what comes after it. An empty string deletes the item.
    If the name is not in the list, then add a new item. }
  procedure THugeStringList.SetValue(const Name, Value: string);
  var
    Equal: Integer;
    I: LongInt;
    Compare: TStrCompareFunc;
  begin
    GetCompareFunc(Compare);
    for I := 0 to Count-1 do
    begin
      Equal := Pos('=', Items[I]^.fString);
      if (Equal <> 0) and (Compare(Name, Copy(Items[I]^.fString, 1, Equal-1)) = 0) then
      begin
        if Value = '' then
          Delete(I)
        else
          Strings[I] := Name + '=' + Value;
        Exit;
      end;
    end;
    if Value <> '' then
      Add(Name + '=' + Value);
  end;

  { Return the index of a string. If the list is sorted, then
    call Find, to use a binary search. }
  function THugeStringList.IndexOf(const S: string): LongInt;
  var
    Compare: TStrCompareFunc;
  begin
    if Sorted then
    begin
      if Find(S, Result) then
        Exit;
    end
    else
    begin
      GetCompareFunc(Compare);
      for Result := 0 to Count-1 do
        if Compare(S, Items[Result]^.fString) = 0 then
          Exit;
    end;
    Result := -1;
  end;

  { Return the index of an object. }
  function THugeStringList.IndexOfObject(Obj: TObject): LongInt;
  begin
    for Result := 0 to Count-1 do
      if Objects[Result] = Obj then
        Exit;
    Result := -1;
  end;

  { Insert at a specific index. Clearly, we cannot do this to a sorted list. }
  procedure THugeStringList.Insert(Index: LongInt; const S: string);
  begin
    Changing;
    if Sorted then
      raise EListError.CreateRes(SSortedListError);
    fList.Insert(Index, NewStritem(S, nil));
    Changed;
  end;

  { Insert a string and an associated object. }
  procedure THugeStringList.InsertObject(Index: LongInt; const S: string; Obj: TObject);
  begin
    Insert(Index, S);
    Objects[Index] := Obj;
  end;

  { Load a new string list from a file }
  procedure THugeStringList.LoadFromFile(const FileName: string);
  var
    Stream: TFileStream;
  begin
    Stream := TFileStream.Create(FileName, fmOpenRead);
    try
      LoadFromStream(Stream);
    finally
      Stream.Free;
    end;
  end;

  { Load a new string list from a stream. Buffer the stream,
    for improved I/O performance.  Line endings separate strings;
    use DOS (CR-LF), Mac (CR), or Unix (LF) style line endings.
    The file ends with a zero byte, a Ctrl+Z, or more commonly,
    reaching the end of the file. }
  procedure THugeStringList.LoadFromStream(Stream: TStream);
  const
    BufSize = 8192;
  var
    Len: Integer;
    Buffer, BufPtr, BufEnd, BufLast: PChar;
    LineLast: PChar;
    S: string;
  begin
    BeginUpdate;
    try
      Clear;
      GetMem(Buffer, BufSize);
      try
        BufEnd := Buffer + BufSize;
        BufLast := BufEnd;
        LineLast := BufLast;
        repeat
          { Copy the last partial line to the start of the buffer }
          Len := BufLast - LineLast;
          if Len <> 0 then
            System.Move(LineLast^, Buffer^, Len);
          BufLast := Buffer + Len;
          { fill the rest of the buffer }
          BufLast := BufLast + Stream.Read(BufLast^, BufEnd - BufLast);
          { find the last end-of-line in the buffer }
          LineLast := BufLast;
          while LineLast > Buffer do
          begin
            Dec(LineLast);
            if LineLast^ in [NUL, LF, CR, EOF] then
              Break;
          end;
          if LineLast^ = CR then
            Inc(LineLast);
          if LineLast^ = LF then
            Inc(LineLast);
          { Now split the buffer into lines }
          BufPtr := Buffer;
          while (BufPtr < LineLast) and (LineLast > Buffer) do
          begin
            BufPtr := ExtractLine(BufPtr, S);
            Add(S);
          end;
        until BufLast = Buffer;
      finally
        FreeMem(Buffer, BufSize);
      end;
    finally
      EndUpdate;
    end;
  end;

  { Move an item from CurIndex to NewIndex. Don't do it if the list is sorted. }
  procedure THugeStringList.Move(CurIndex, NewIndex: LongInt);
  begin
    if Sorted then
      raise EListError.CreateRes(SSortedListError);
    Changing;
    fList.Move(CurIndex, NewIndex);
    Changed;
  end;

  { Parition for quick sort. Override this if you want to use a more
    sophisticated algorithm for finding the pivot. }
  function THugeStringList.Partition(Lo, Hi: LongInt; Compare: TStrCompareFunc): LongInt;
  var
    Pivot: PStrItem;
  begin
    { Choose a pivot.  We can do a better job, but this will suffice for now. }
    Pivot := fList[(Lo + Hi) div 2];
    while true do
    begin
      while Compare(PStrItem(fList[Lo])^.fString, Pivot^.fString) < 0 do
        Inc(Lo);
      while Compare(PStrItem(fList[Hi])^.fString, Pivot^.fString) > 0 do
        Dec(Hi);
      if Lo < Hi then
      begin
        fList.Exchange(Lo, Hi);
        Inc(Lo);
        Dec(Hi);
      end
      else if Lo >= Hi then
        Break;
    end;
    Result := Hi;
  end;

  { Slow sort for small lists. }
  procedure THugeStringList.SlowSort(Lo, Hi: LongInt; Compare: TStrCompareFunc);
  var
    I, J: LongInt;
  begin
    for I := Lo to Hi do
      for J := I+1 to Hi do
        if Compare(Items[I]^.fString, Items[J]^.fString) > 0 then
          fList.Exchange(I, J);
  end;

  { Quick sort. Use tail recursion, to minimize the stack growth.
    Use a slow sort for small lists. The name, "SlowSort," is merely an
    indicator that of its asymptotic performance. Typically, SlowSort is
    a bubble sort or insertion sort, but one can override it for a Shell sort,
    for example. Hence, the generic name. }
  procedure THugeStringList.QuickSort(Lo, Hi: LongInt; Compare: TStrCompareFunc);
  var
    Mid: LongInt;
    I, L, H: LongInt;
  begin
    L := Lo;
    H := Hi;
    { Force tail recursion to sort the larger partition. This keeps the stack
      size from growing too much, and helps the speed a little. }
    while Lo < Hi do
      if Hi-Lo < SortThreshold then
      begin
        SlowSort(Lo, Hi, Compare);
        Break;
      end
      else
      begin
        Mid := Partition(Lo, Hi, Compare);
        if Mid-Lo > Hi-Mid then
        begin
          QuickSort(Mid+1, Hi, Compare);
          Hi := Mid;
        end
        else
        begin
          QuickSort(Lo, Mid, Compare);
          Lo := Mid+1;
        end;
      end;
  end;

  { Save all the strings to a file. }
  procedure THugeStringList.SaveToFile(const FileName: string);
  var
    Stream: TFileStream;
  begin
    Stream := TFileStream.Create(FileName, fmCreate);
    try
      SaveToStream(Stream);
    finally
      Stream.Free;
    end;
  end;

  { Save all the strings to a stream, one per line. }
  procedure THugeStringList.SaveToStream(Stream: TStream);
  const
    BufSize = 8192;
  var
    I: LongInt;
    Buffer, BufPtr: PChar;
  begin
    GetMem(Buffer, BufSize);
    try
      BufPtr := Buffer;
      for I := 0 to Count-1 do
      begin
        { Is there room for a string + CRLF? }
        if BufPtr - Buffer >= BufSize - 256 then
        begin
          Stream.WriteBuffer(Buffer[0], BufPtr - Buffer);
          BufPtr := Buffer;
        end;
        StrPCopy(BufPtr, Items[I]^.fString);
        Inc(BufPtr, Length(Items[I]^.fString));
        BufPtr[0] := CR;
        BufPtr[1] := LF;
        Inc(BufPtr, 2);
      end;
      Stream.WriteBuffer(Buffer[0], BufPtr - Buffer);
    finally
      FreeMem(Buffer, BufSize);
    end;
  end;

  { Set all the strings by reading lines from Text. }
  procedure THugeStringList.SetText(Text: PChar);
  var
    S: string;
  begin
    BeginUpdate;
    try
      Clear;
      while not (Text^ in [NUL, EOF]) do
      begin
        Text := ExtractLine(Text, S);
        Add(S);
      end;
    finally
      EndUpdate;
    end;
  end;

  { Sort the list, if it is not already sorted. }
  procedure THugeStringList.Sort;
  var
    Compare: TStrCompareFunc;
  begin
    if not Sorted and (Count > 1) then
    begin
      GetCompareFunc(Compare);
      Changing;
      QuickSort(0, Count-1, Compare);
      Changed;
    end;
  end;

  { THugeStrings }
  { In many cases, we can hide the long indexes, but sometimes,
    we cannot, e.g., IndexOf. In these situations, we let Delphi
    catch an overflow error. }
  {$Q+}
  constructor THugeStrings.Create;
  begin
    fList := THugeStringList.Create;
    inherited Create;
  end;

  destructor THugeStrings.Destroy;
  begin
    fList.Free;
    inherited Destroy;
  end;

  function THugeStrings.GetDuplicates: TDuplicates;
  begin
    Result := fList.Duplicates
  end;

  function THugeStrings.GetSorted: Boolean;
  begin
    Result := fList.Sorted
  end;

  function THugeStrings.GetOnChange: TNotifyEvent;
  begin
    Result := fList.OnChange
  end;

  function THugeStrings.GetOnChanging: TNotifyEvent;
  begin
    Result := fList.OnChanging
  end;

  procedure THugeStrings.SetOnChange(OnChange: TNotifyEvent);
  begin
    fList.OnChange := OnChange;
  end;

  procedure THugeStrings.SetOnChanging(OnChanging: TNotifyEvent);
  begin
    fList.OnChanging := OnChanging
  end;

  procedure THugeStrings.SetDuplicates(Duplicates: TDuplicates);
  begin
    fList.Duplicates := Duplicates
  end;

  procedure THugeStrings.SetSorted(Sorted: Boolean);
  begin
    fList.Sorted := Sorted
  end;

  function THugeStrings.Get(Index: Integer): string;
  begin
    Result := fList[Index]
  end;

  function THugeStrings.GetCount: Integer;
  begin
    Result := Trunc(fList.Count);
  end;

  function THugeStrings.HugeGetCount: LongInt;
  begin
    Result := fList.Count
  end;

  function THugeStrings.GetObject(Index: Integer): TObject;
  begin
    Result := fList.Objects[Index]
  end;

  procedure THugeStrings.Put(Index: Integer; const S: string);
  begin
    fList[Index] := S
  end;

  procedure THugeStrings.PutObject(Index: Integer; Obj: TObject);
  begin
    fList.Objects[Index] := Obj
  end;

  procedure THugeStrings.SetUpdateState(Updating: Boolean);
  begin
    if Updating then
      fList.BeginUpdate
    else
      fList.EndUpdate
  end;

  function THugeStrings.HugeGet(Index: LongInt): string;
  begin
    Result := fList[Index]
  end;

  function THugeStrings.HugeGetObject(Index: LongInt): TObject;
  begin
    Result := fList.Objects[Index]
  end;

  procedure THugeStrings.HugePut(Index: LongInt; const S: string);
  begin
    fList[Index] := S
  end;

  procedure THugeStrings.HugePutObject(Index: LongInt; Obj: TObject);
  begin
    fList.Objects[Index] := Obj
  end;

  function THugeStrings.HugeGetValue(const Name: string): string;
  begin
    Result := fList.Values[Name]
  end;

  procedure THugeStrings.HugeSetValue(const Name, Value: string);
  begin
    fList.Values[Name] := Value
  end;

  procedure THugeStrings.ReadData(Reader: TReader);
  begin
    fList.ReadData(Reader)
  end;

  procedure THugeStrings.WriteData(Writer: TWriter);
  begin
    fList.WriteData(Writer)
  end;

  procedure THugeStrings.DefineProperties(Filer: TFiler);
  begin
    Filer.DefineProperty('Strings', ReadData, WriteData, HugeCount>1)
  end;

  function THugeStrings.Add(const S: string): Integer;
  begin
    Result := Trunc(fList.Add(S))
  end;

  function THugeStrings.HugeAdd(const S: string): LongInt;
  begin
    Result := fList.Add(S)
  end;

  function THugeStrings.AddObject(const S: string; Obj: TObject): Integer;
  begin
    Result := Trunc(fList.AddObject(S, Obj));
  end;

  function THugeStrings.HugeAddObject(const S: string; Obj: TObject): LongInt;
  begin
    Result := fList.AddObject(S, Obj);
  end;

  procedure THugeStrings.Assign(Source: TPersistent);
  begin
    fList.Assign(Source)
  end;

  procedure THugeStrings.Clear;
  begin
    fList.Clear
  end;

  procedure THugeStrings.Delete(Index: Integer);
  begin
    fList.Delete(Index)
  end;

  procedure THugeStrings.HugeDelete(Index: LongInt);
  begin
    fList.Delete(Index)
  end;

  procedure THugeStrings.Exchange(Index1, Index2: Integer);
  begin
    fList.Exchange(Index1, Index2)
  end;

  procedure THugeStrings.HugeExchange(Index1, Index2: LongInt);
  begin
    fList.Exchange(Index1, Index2)
  end;

  function THugeStrings.Find(const S: string; var Index: Integer): Boolean;
  var
    HugeIndex: LongInt;
  begin
    Result := fList.Find(S, HugeIndex);
    Index := Trunc(HugeIndex);
  end;

  function THugeStrings.HugeFind(const S: string; var Index: LongInt): Boolean;
  begin
    Result := fList.Find(S, Index);
  end;

  function THugeStrings.GetText: PChar;
  begin
    Result := fList.GetText;
  end;

  function THugeStrings.IndexOf(const S: string): Integer;
  begin
    Result := fList.IndexOf(S)
    { Let Delphi catch an overflow error if the Index > MaxInt }
  end;

  function THugeStrings.HugeIndexOf(const S: string): LongInt;
  begin
    Result := fList.IndexOf(S)
  end;

  function THugeStrings.HugeIndexOfObject(Obj: TObject): LongInt;
  begin
    Result := fList.IndexOfObject(Obj)
  end;

  procedure THugeStrings.Insert(Index: Integer; const S: string);
  begin
    fList.Insert(Index, S)
  end;

  procedure THugeStrings.HugeInsert(Index: LongInt; const S: string);
  begin
    fList.Insert(Index, S)
  end;

  procedure THugeStrings.HugeInsertObject(Index: LongInt; const S: string; Obj: TObject);
  begin
    fList.InsertObject(Index, S, Obj)
  end;

  procedure THugeStrings.LoadFromStream(Stream: TStream);
  begin
    fList.LoadFromStream(Stream)
  end;

  procedure THugeStrings.Move(CurIndex, NewIndex: Integer);
  begin
    fList.Move(CurIndex, NewIndex);
  end;

  procedure THugeStrings.HugeMove(CurIndex, NewIndex: LongInt);
  begin
    fList.Move(CurIndex, NewIndex)
  end;

  procedure THugeStrings.SaveToStream(Stream: TStream);
  begin
    fList.SaveToStream(Stream)
  end;

  procedure THugeStrings.SetText(Text: PChar);
  begin
    fList.SetText(Text)
  end;

  procedure THugeStrings.Sort;
  begin
    fList.Sort
  end;

  function THugeStrings.Trunc(Index: LongInt): Integer;
  begin
    if Index > MaxInt then
      Result := MaxInt
    else
      Result := Integer(Index)
  end;

  procedure Register;
  begin
    Classes.RegisterClasses([THugeStringList, THugeStrings]);
  end;

end.
{--=>revision-history<=--}
{1 HUGE.PAS 7-Jun-95,15:12:08,`lisch' CHANGE=INITIAL                          }
{     Initial version.                                                        }
{2 HUGE.PAS 22-Aug-95,13:41:06,`lisch' CHANGE=V14                             }
{     Fixed Move to delete & insert, instead of Copy.                         }
{3 HUGE.PAS 3-Sep-95,15:07:58,`lisch' CHANGE=V14                              }
{     Version 1.4: Optimized Move.  Added Move/IndexOf to demo program.       }
{4 HUGE.PAS 17-Sep-95,20:03:20,`lisch' CHANGE=STRLIST                         }
{     Added THugeStringList                                                   }
{5 HUGE.PAS 16-Oct-95,14:47:36,`lisch' CHANGE=STRINGLIST                      }
{     Added THugeStringList, et al.                                           }
{     Also improved THugeList by creating a custom base class                 }
{6 HUGE.PAS 11-Sep-98,18:02:03,'sam'
{     Changed HUGEMOVE so it clears the direction flag (CLD) before returning }
{ Modification made 1998 S Liddicott, search for 'SAMFIX'. samjam@bigfoot.com }
{--=>revision-history<=--}
